//standard
#include <cstdlib>
#include <cstdio>	//too used to printf
#include <vector>
#include <cmath>
#include <cstring>
#include <iostream>

#include <inttypes.h>

//system
#include <unistd.h>
#include <sys/types.h>	//for getpid: http://linux.die.net/man/2/getpid
#include <libgen.h>
#include <dirent.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/statfs.h>
#include <errno.h>
#include <sys/wait.h>

//gtk
#include <gtk/gtk.h>
#include <gtk/gtkimagemenuitem.h>

//gvfs/gio
#include <gio/gio.h>

//xml
#include <tinyxml/tinyxml.h>

//curl
#include <curl/curl.h>

//regex
#include <pcre.h>

//bz2
#include <bzlib.h>

//tar
#include <libtar.h>

//set the output functions
#ifdef DEBUG
	//use printf for 
	#define DBG_PRINT		printf
#else
	//gobble the debug up if DEBUG isn't set
	#define	DBG_PRINT(...)
#endif
//redirect to stderr
#define ERROR_PRINT(...)	fprintf(stderr, __VA_ARGS__)

//merge as directory and file
char* dir_merge(const char* arg1, const char* arg2)
{
	char* ret = new char[strlen(arg1) + strlen(arg2) + 2];
	char* slash = strrchr(arg1, '/');	//DO NOT DELETE
	if(slash && slash == arg1 + strlen(arg1) - 1)
		strcat(strcpy(ret, arg1), arg2);
	else
		strcat(strcat(strcpy(ret, arg1),  "/"), arg2);
	return(ret);
}

//is this a directory?
int is_dir(const char* fname)
{
	//using open and fstat
	int fd = open(fname, O_RDONLY);
	struct stat *file_stat = new(struct stat);
	int exists = !fstat(fd, file_stat);	//exists?
	close(fd);
	if(exists)
	{
		int ret = S_ISDIR(file_stat->st_mode);
		delete file_stat;
		return(ret);
	}
	delete file_stat;
	return(0);
}

static GtkListStore* store;
static GtkWidget* tree;
static char* current_path = NULL;
enum{ NAME_COL, PATH_COL, NCOLUMNS };

/*
//adapted from http://zetcode.com/tutorials/gtktutorial/gtktreeview/
void on_changed(GtkWidget *widget, gpointer label) 
{
	GtkTreeIter iter;
	GtkTreeModel *model;

	if(gtk_tree_selection_get_selected(GTK_TREE_SELECTION(widget), &model, &iter))
	{
		gtk_tree_model_get(model, &iter, PATH_COL, &current_path,  -1);
		DBG_PRINT("Firefox Config: selection item with path \"%s\"\n", current_path);
	}
}
// */


extern "C" char* make_config_widget(GtkWidget** save_to)
{
	*save_to = gtk_vbox_new(FALSE, 2);
	
	GtkWidget* instruction_label = gtk_label_new("Choose a Firefox profile below:");
	gtk_container_add(GTK_CONTAINER(*save_to), instruction_label);
	
	GtkWidget* list_window = gtk_scrolled_window_new(NULL, NULL);
	
	//adapted from http://zetcode.com/tutorials/gtktutorial/gtktreeview/
	
	GtkTreeIter iter;

	store = gtk_list_store_new(NCOLUMNS, G_TYPE_STRING, G_TYPE_STRING);
	tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (store));
	GtkCellRenderer* renderer = gtk_cell_renderer_text_new();
	GtkTreeViewColumn* name = gtk_tree_view_column_new_with_attributes("Profiles", renderer, "text", NAME_COL, NULL);
	GtkTreeViewColumn* path = gtk_tree_view_column_new_with_attributes("Path", renderer, "text", PATH_COL, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree), name);
	gtk_tree_view_append_column(GTK_TREE_VIEW(tree), path);
	gtk_tree_view_set_model(GTK_TREE_VIEW(tree), GTK_TREE_MODEL(store));
	
	//g_signal_connect(selection, "changed", G_CALLBACK(on_changed), label);

	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(list_window), tree);
	gtk_container_add(GTK_CONTAINER(*save_to), list_window);
	
	char* dirname = dir_merge(getenv("HOME"), ".mozilla/firefox");
	DBG_PRINT("Firefox Config: dirname: %s\n", dirname);
	
	int matches[99];
	static const char* error;
	static int erroffset;
	//Matches: 8 alpha numerics + a period + everything else, capture the last one
	static pcre* dir_regex = pcre_compile("\\w{8}\\.(.*)", 0, &error, &erroffset, NULL);
	//if the directory exists
	if(is_dir(dirname))
	{
		//get the directory to list the files
		DIR* dir = opendir(dirname);
		//get the structure in the files would be read
		dirent* each_ent;
		//the loop variables:
		char* each_path;
		char* each_name;
		GtkWidget* each_label;
		//go one at a time...
		while((each_ent = readdir(dir)) != NULL)
		{
			//the directory isn't "." or ".."
			if(strcmp(each_ent->d_name, ".") && strcmp(each_ent->d_name, ".."))
			{
				//find the absolute path of the folder
				each_path = dir_merge(dirname, each_ent->d_name);
				DBG_PRINT("Firefox Config: each_name: %s\n", each_path);
				int result = pcre_exec(dir_regex, NULL, each_ent->d_name, strlen(each_ent->d_name), 0, 0, matches, 99);
				//does it match the regex, and is it a directory?
				if(result >= 2 && is_dir(each_path))
				{
					each_name = new char[matches[3] - matches[2] + 1];
					sprintf(each_name, "%.*s", (matches[3] - matches[2]), (each_ent->d_name + matches[2]));
					//yes, move on...
					DBG_PRINT("Firefox Config: \tis_dir; name: %s\n", each_name);
					gtk_list_store_append(store, &iter);
					gtk_list_store_set(store, &iter, NAME_COL, each_name, PATH_COL, each_path, -1);
					delete[] each_name;
				}
				//clean up even if it wasn't a folder
				delete[] each_path;
			}
		}
		//finally, close it; otherwise, I can unmount the media
		closedir(dir);
	}
	else
	{
		ERROR_PRINT("Firefox Config: Couldn't find configuration directory (HOME/.mozilla/firefox)!\n");
		return("Couldn't find configuration directory (HOME/.mozilla/firefox)!");
	}
	
	//normal exit
	return(NULL);
}

extern "C" char* do_config(char* path)
{
	DBG_PRINT("Firefox Config: coping to path \"%s\"\n", path);
	//*
	GtkTreeIter iter;
	GtkTreeModel* model;
	gtk_tree_selection_get_selected(GTK_TREE_SELECTION(gtk_tree_view_get_selection(GTK_TREE_VIEW(tree))), &model, &iter);
	char* current_path = NULL;
	gtk_tree_model_get(model, &iter, PATH_COL, &current_path,  -1);
	DBG_PRINT("Firefox Config: selection item with path \"%s\"\n", current_path);
	if(current_path != NULL)
	{
		char* profile_folder = dir_merge(path, "profile");
		if(mkdir(profile_folder, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) == 0)
		{
			char* cmd = new char[strlen(profile_folder) + strlen(current_path) + 1024];
			sprintf(cmd, "cp -r '%s'/* '%s'", current_path, profile_folder);
			system(cmd);
			DBG_PRINT("Firefox Config: copy command: %s\n", cmd);
			delete[] cmd;
		}
		else
		{
			return("Couldn't make profile folder!");
			delete[] profile_folder;
		}
		delete[] profile_folder;
	}
	else
	{
		return("Couldn't get profile folder!");
	}
	// */
	return(NULL);
}
